
/*
 *  main.c
 * Copyright  1992 Niklas Rjemo
 */
#include <setjmp.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "strdup.h"
#include "input.h"
#include "output.h"
#include "error.h"
#include "asm.h"
#include "area.h"
#include "riscos.h"
#include "include.h"
#include "variables.h"
#include "version.h"
#include "targetcpu.h"


jmp_buf asmContinue;
jmp_buf asmAbort;

int verbose=0;
int pedantic=0;
int throwback=0;
int dde=0;
int autocast=0;
int align=1;
int gcc_backend=0;
int gas_backend=0;
int local=1;
int objasm=0;
int uc=0;

char *ProgName;
char *ObjFileName;
char *SourceFileName;

static void as_help(char *progname)
{
  fprintf(stderr,
    "%s"
    "\n"
    "Usage: %s [option]... <asmfile>\n"
    "\n"
    "Options:\n"
    "-o objfile        Destination AOF file\n"
    "-Idirectory       Search 'directory' for included assembler files\n"
    "-pedantic   -p    Enable pedantic mode\n"
    "-verbose    -v    Give information about the progress\n"
    "-dde              Replace '@' in filenames with <Prefix$Dir>\n"
    "-throwback  -tb   Throwback errors to a text editor\n"
    "-autocast   -ac   Enable casting from integer to float\n"
    "-target     -t    Target ARM CPU (ARM2...SA110)\n"
    "-gcc              Predefine register numbers for GCC\n"
    "-gas              Support some GNU GAS keywords\n"
    "-noalign    -na   Don't auto-align words and halfwords\n"
    "-nolocal    -nl   No builtin LOCAL support\n"
    "-objasm     -obj  More compatibility with ObjAsm\n"
    "-upper      -up   Mnemonics must be in upper case\n"
    "-help -h -H -?    Display this help\n"
    "-version    -ver  Display the version number\n"
    "-From asmfile     Source assembler file (Objasm compatibility)\n"
    "-To objfile       Destination AOF file (Objasm compatibility)\n"
    "\n",
    VERSION, progname);
}


extern void as_target(char*); /* in targetcpu.c */
extern int _kernel_setenv(const char *, const char *);

static char *prefix;
static int finished=0;


static void restore_prefix(void)
{
  if (prefix) _kernel_setenv("Prefix$Dir",prefix);
  if (!finished || noerrors()) outputRemove();
    /* workaround for throwback/Prefix$Dir problem */
}


int main(int argc, char **argv)
{
  ProgName=getenv("Prefix$Dir");
   /* There's a strange problem with Prefix$Dir becoming unset if
    * throwback is used...
    */
  if (ProgName) {
    prefix=strdup(ProgName);
    if (prefix) atexit(restore_prefix);
  }

    initInclude();
  ProgName = *argv++;
  ObjFileName = NULL;

#define IS_ARG(ln,sn) !strcmp(*argv,ln) || !strcmp(*argv,sn)

  if (argc == 1)
    {
      /* No command line arguments supplied. Print help and exit.  */
      as_help (ProgName);
      return 0;
    }

  for (argc--; argc; argv++, argc--) {
    if (argv[0][0]=='-' && argv[0][1]=='D') {
      if (argv[0][2]==0) {
        if (--argc)
          var_define(*++argv);
        else {
          fprintf(stderr,"%s: Missing filename after -o\n",ProgName);
          return -1;
        }
      } else
        var_define(argv[0]+2);
    } else if (IS_ARG("-o","-To")) {
      if (ObjFileName) {
        fprintf(stderr,"%s: Only one output file allowed\n",ProgName);
        return -1;
      }
      if (--argc)
        ObjFileName = *++argv;
      else {
        fprintf(stderr,"%s: Missing filename after -o\n",ProgName);
        return -1;
      }
    } else if (IS_ARG("-throwback","-tb")) throwback++;
    else if (IS_ARG("-autocast","-ac")) autocast++;
    else if (IS_ARG("-noalign","-na")) align=0;
    else if (IS_ARG("-nolocal","-nl")) local=0;
    else if (IS_ARG("-objasm","-obj")) objasm=1;/* used as index in c.lex* */
    else if (IS_ARG("-upper","-up")) uc++;
    else if (IS_ARG("-pedantic","-p")) pedantic++;
    else if (IS_ARG("-target","-t")) as_target(--argc ? *++argv : 0);
    else if (IS_ARG("-verbose","-v")) verbose++;
    else if (!strcmp(*argv,"-gcc")) gcc_backend++;
    else if (!strcmp(*argv,"-gas")) gas_backend++;
    else if (!strcmp(*argv,"-dde")) dde++;
    else if (!strncmp(*argv,"-I",2)) {
      if (addInclude(*argv+2)<0) return 1;
    } else if (IS_ARG("-version","-ver")) {
      fprintf(stderr, "%s%s", VERSION, COPYRIGHT);
      return 0;
    } else if (IS_ARG("-H","-h") || IS_ARG("-help","-?")) {
      as_help(ProgName);
      return 0;
    } else if (!strcmp(*argv,"-From")) {
      if (SourceFileName) {
        fprintf(stderr,"%s: Only one input file allowed\n",ProgName);
        return -1;
      }
      if (--argc)
        SourceFileName = *++argv;
      else {
        fprintf(stderr,"%s: Missing filename after -From\n",ProgName);
        return -1;
      }
    } else if (**argv != '-') {
      if (SourceFileName) {
        fprintf(stderr,"%s: Only one input file allowed\n",ProgName);
        return -1;
      }
      SourceFileName = *argv;
    } else
      fprintf(stderr,"%s: Illegal flag %s ignored\n",ProgName,*argv);
  }
  if (!ObjFileName) ObjFileName = SourceFileName;

  if (setjmp(asmAbort)) {
    fprintf(stderr,"%s: Aborted\n",ProgName);
    inputFinish();
    outputFinish();
  } else {
    inputInit(SourceFileName);
    errorInit(throwback, inputName);
    outputInit(ObjFileName);
    areaInit();
    setjmp(asmContinue);
    asm_();
    areaFinish();
    inputFinish();
    if (setjmp(asmContinue))
      fprintf(stdout,"%s: Error when writing object file '%s'.\n",ProgName,ObjFileName);
    else
      outputAof();
    outputFinish();
    errorFinish();
  }
  finished=1;
  return noerrors();
}
